@@ -147,7 +147,8 @@ static void     gdm_slave_usr2_handler (int sig);
 static void     gdm_slave_exit (gint status, const gchar *format, ...) G_GNUC_PRINTF (2, 3);
 static void     gdm_child_exit (gint status, const gchar *format, ...) G_GNUC_PRINTF (2, 3);
 static gint     gdm_slave_exec_script (GdmDisplay *d, const gchar *dir,
-				       const char *login, struct passwd *pwent);
+				       const char *login, struct passwd *pwent,
+				       gboolean pass_stdout);
 static gchar *  gdm_parse_enriched_login (const gchar *s, GdmDisplay *display);
 static void	gdm_slave_handle_notify (const char *msg);
 
@@ -319,7 +320,8 @@ setup_automatic_session (GdmDisplay *display, const char *name)
 
 	/* Run the init script. gdmslave suspends until script
 	 * has terminated */
-	gdm_slave_exec_script (display, GdmDisplayInit, NULL, NULL);
+	gdm_slave_exec_script (display, GdmDisplayInit, NULL, NULL,
+			       FALSE /* pass_stdout */);
 
 	gdm_debug ("setup_automatic_session: DisplayInit script finished");
 
@@ -393,7 +395,7 @@ gdm_slave_whack_greeter (void)
 	/* Wait for the greeter to really die, the check is just
 	 * being very anal, the pid is always set to something */
 	if (d->greetpid > 0)
-		waitpid (d->greetpid, 0, 0); 
+		ve_waitpid_no_signal (d->greetpid, 0, 0); 
 	d->greetpid = 0;
 
 	gdm_slave_send_num (GDM_SOP_GREETPID, 0);
@@ -502,6 +504,16 @@ gdm_slave_run (GdmDisplay *display)
 	}
     }
 
+    /* Set the busy cursor */
+    if (d->dsp != NULL) {
+	    Cursor xcursor = XCreateFontCursor (d->dsp, GDK_WATCH);
+	    XDefineCursor (d->dsp,
+			   DefaultRootWindow (d->dsp),
+			   xcursor);
+	    XFreeCursor (d->dsp, xcursor);
+	    XSync (d->dsp, False);
+    }
+
     /* Just a race avoiding sleep, probably not necessary though,
      * but doesn't hurt anything */
     if ( ! d->handled)
@@ -654,7 +666,6 @@ focus_first_x_window (const char *class_res_name)
 	pid_t pid;
 	Display *disp;
 	XWindowAttributes attribs = { 0, };
-	int i;
 
 	pid = fork ();
 	if (pid < 0) {
@@ -668,8 +679,7 @@ focus_first_x_window (const char *class_res_name)
 
 	closelog ();
 
-	for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-	    close(i);
+	gdm_close_all_descriptors (0 /* from */, -1 /* except */);
 
 	/* No error checking here - if it's messed the best response
          * is to ignore & try to continue */
@@ -748,8 +758,18 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 		display->sesspid = 0;
 		return;
 	}
+
+	/* Set the busy cursor */
+	if (d->dsp != NULL) {
+		Cursor xcursor = XCreateFontCursor (d->dsp, GDK_WATCH);
+		XDefineCursor (d->dsp,
+			       DefaultRootWindow (d->dsp),
+			       xcursor);
+		XFreeCursor (d->dsp, xcursor);
+		XSync (d->dsp, False);
+	}
+
 	if (pid == 0) {
-		int i;
 		char **argv;
 		/* child */
 
@@ -774,8 +794,7 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 
 		closelog ();
 
-		for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-			close(i);
+		gdm_close_all_descriptors (0 /* from */, -1 /* except */);
 
 		/* No error checking here - if it's messed the best response
 		 * is to ignore & try to continue */
@@ -821,7 +840,7 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 			/* must use the pid var here since sesspid might get
 			 * zeroed between the check and here by sigchld
 			 * handler */
-			waitpid (pid, 0, 0);
+			ve_waitpid_no_signal (pid, 0, 0);
 		display->sesspid = 0;
 	}
 }
@@ -837,12 +856,14 @@ restart_the_greeter (void)
 	if (greet) {
 		gdm_sigchld_block_push ();
 
+		gdm_slave_greeter_ctl_no_ret (GDM_SAVEDIE, "");
+
 		greet = FALSE;
 
-		if (d->greetpid > 0) {
-			if (kill (d->greetpid, SIGTERM) == 0)
-				waitpid (d->greetpid, 0, 0); 
-		}
+		/* Wait for the greeter to really die, the check is just
+		 * being very anal, the pid is always set to something */
+		if (d->greetpid > 0)
+			ve_waitpid_no_signal (d->greetpid, 0, 0); 
 		d->greetpid = 0;
 
 		gdm_slave_send_num (GDM_SOP_GREETPID, 0);
@@ -1234,13 +1255,12 @@ run_pictures (void)
 		}
 		g_free (ret);
 
-		printf ("%c", STX);
-		fflush (stdout);
+		gdm_fdprintf (STDOUT_FILENO, "%c", STX);
 
 		i = 0;
 		while ((bytes = fread (buf, sizeof (char),
 				       sizeof (buf), fp)) > 0) {
-			fwrite (buf, 1, bytes, stdout);
+			write (STDOUT_FILENO, buf, bytes);
 			i += bytes;
 		}
 
@@ -1249,7 +1269,7 @@ run_pictures (void)
 		/* eek, this "could" happen, so just send some garbage */
 		while (i < s.st_size) {
 			bytes = MIN (sizeof (buf), s.st_size - i);
-			fwrite (buf, 1, bytes, stdout);
+			write (STDOUT_FILENO, buf, bytes);
 			i += bytes;
 		}
 			
@@ -1267,13 +1287,13 @@ gdm_slave_greeter (void)
     gint pipe1[2], pipe2[2];  
     gchar **argv;
     struct passwd *pwent;
-    int i;
     pid_t pid;
     
     gdm_debug ("gdm_slave_greeter: Running greeter on %s", d->name);
     
     /* Run the init script. gdmslave suspends until script has terminated */
-    gdm_slave_exec_script (d, GdmDisplayInit, NULL, NULL);
+    gdm_slave_exec_script (d, GdmDisplayInit, NULL, NULL,
+			   FALSE /* pass_stdout */);
 
     /* Open a pipe for greeter communications */
     if (pipe (pipe1) < 0 || pipe (pipe2) < 0) 
@@ -1309,8 +1329,7 @@ gdm_slave_greeter (void)
 
 	closelog ();
 
-	for (i = 2; i < sysconf (_SC_OPEN_MAX); i++)
-	    close(i);
+	gdm_close_all_descriptors (2 /* from */, -1 /* except */);
 
 	open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
 
@@ -1461,14 +1480,15 @@ gdm_slave_greeter (void)
 	/* flush our input before we change the piping */
 	fflush (stdin);
 
-	if (pipe1[1] != STDOUT_FILENO) 
+	if (pipe1[1] != STDOUT_FILENO) {
 	    dup2 (pipe1[1], STDOUT_FILENO);
+	    close (pipe1[1]);
+	}
 	
-	if (pipe2[0] != STDIN_FILENO) 
+	if (pipe2[0] != STDIN_FILENO)  {
 	    dup2 (pipe2[0], STDIN_FILENO);
-
-	close (pipe1[1]);
-	close (pipe2[0]);
+	    close (pipe2[0]);
+	}
 	
 	gdm_debug ("gdm_slave_greeter: Greeter on pid %d", (int)pid);
 
@@ -1622,7 +1642,6 @@ gdm_slave_chooser (void)
 	struct passwd *pwent;
 	char buf[1024];
 	size_t bytes;
-	int i;
 	pid_t pid;
 
 	gdm_debug ("gdm_slave_chooser: Running chooser on %s", d->name);
@@ -1632,7 +1651,8 @@ gdm_slave_chooser (void)
 		gdm_slave_exit (DISPLAY_ABORT, _("gdm_slave_chooser: Can't init pipe to gdmchooser"));
 
 	/* Run the init script. gdmslave suspends until script has terminated */
-	gdm_slave_exec_script (d, GdmDisplayInit, NULL, NULL);
+	gdm_slave_exec_script (d, GdmDisplayInit, NULL, NULL,
+			       FALSE /* pass_stdout */);
 
 	/* Fork. Parent is gdmslave, child is greeter process. */
 	gdm_sigchld_block_push ();
@@ -1659,8 +1679,7 @@ gdm_slave_chooser (void)
 		closelog ();
 
 		close (0);
-		for (i = 2; i < sysconf (_SC_OPEN_MAX); i++)
-			close(i);
+		gdm_close_all_descriptors (2 /* from */, -1 /* except */);
 
 		open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
 		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
@@ -1735,7 +1754,7 @@ gdm_slave_chooser (void)
 		gdm_sigchld_block_push ();
 		/* wait for the chooser to die */
 		if (d->chooserpid > 0)
-			waitpid (d->chooserpid, 0, 0);
+			ve_waitpid_no_signal (d->chooserpid, 0, 0);
 		d->chooserpid  = 0;
 		gdm_sigchld_block_pop ();
 
@@ -2050,13 +2069,14 @@ session_child_run (struct passwd *pwent,
 {
 	sigset_t mask; 
 	int logfd;
-	int i;
 	gboolean failsafe = FALSE;
 	char *sesspath, *sessexec;
 	gboolean need_config_sync = FALSE;
 	const char *shell = NULL;
 	Display *disp;
 
+	gnome_setenv ("XAUTHORITY", d->authfile, TRUE);
+
 	disp = XOpenDisplay (d->name);
 	if (disp != NULL) {
 		XSetInputFocus (disp, PointerRoot,
@@ -2064,6 +2084,53 @@ session_child_run (struct passwd *pwent,
 		XCloseDisplay (disp);
 	}
 
+	if (strcmp (session, GDM_SESSION_FAILSAFE_GNOME) == 0 ||
+	    strcmp (session, GDM_SESSION_FAILSAFE_XTERM) == 0 ||
+	    /* hack */
+	    g_ascii_strcasecmp (session, "Failsafe") == 0) {
+		failsafe = TRUE;
+	}
+
+	/* Here we setup our 0,1,2 descriptors, we do it here
+	 * nowdays rather then later on so that we get errors even
+	 * from the PreSession script */
+        /* Log all output from session programs to a file,
+	 * unless in failsafe mode which needs to work when there is
+	 * no diskspace as well */
+	if ( ! failsafe && home_dir_ok) {
+		logfd = open (g_strconcat (home_dir, "/.xsession-errors", NULL),
+			      O_CREAT|O_TRUNC|O_WRONLY, 0644);
+		if (logfd != -1) {
+			dup2 (logfd, 1);
+			dup2 (logfd, 2);
+			close (logfd);
+		} else {
+			close (1);
+			close (2);
+			open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
+			open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+			gdm_error (_("%s: Could not open ~/.xsession-errors"),
+				   "run_session_child");
+		}
+	} else {
+		close (1);
+		close (2);
+		open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
+		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+	}
+
+	close (0);
+	open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+
+	/* Run the PreSession script */
+	if (gdm_slave_exec_script (d, GdmPreSession,
+				   login, pwent,
+				   TRUE /* pass_stdout */) != EXIT_SUCCESS &&
+	    /* ignore errors in failsafe modes */
+	    ! failsafe) 
+		/* If script fails reset X server and restart greeter */
+		gdm_child_exit (DISPLAY_REMANAGE,
+				_("gdm_slave_session_start: Execution of PreSession script returned > 0. Aborting."));
 
 	gdm_clearenv ();
 
@@ -2130,8 +2197,11 @@ session_child_run (struct passwd *pwent,
 		gnome_setenv ("LANG", language, TRUE);
 		gnome_setenv ("GDM_LANG", language, TRUE);
 	} else {
-		/* setusercontext sets up user languages */
-		gnome_setenv ("GDM_LANG", g_getenv ("LANG"), TRUE);
+		if (g_getenv ("LANG") == NULL)
+			gnome_unsetenv ("GDM_LANG");
+		else
+			/* setusercontext sets up user languages */
+			gnome_setenv ("GDM_LANG", g_getenv ("LANG"), TRUE);
 	}
 #else
 	if (setuid (pwent->pw_uid) < 0) 
@@ -2160,7 +2230,11 @@ session_child_run (struct passwd *pwent,
 	if (usrcfgok && savelang && home_dir_ok) {
 		gchar *cfgstr = g_strconcat ("=", home_dir,
 					     "/.gnome/gdm=/session/lang", NULL);
-		gnome_config_set_string (cfgstr, language);
+		/* we chose the system default language so wipe the lang key */
+		if (def_language)
+			gnome_config_clean_key (cfgstr);
+		else
+			gnome_config_set_string (cfgstr, language);
 		need_config_sync = TRUE;
 		g_free (cfgstr);
 	}
@@ -2182,14 +2256,7 @@ session_child_run (struct passwd *pwent,
 
 	closelog ();
 
-	for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-	    close (i);
-
-	/* No error checking here - if it's messed the best response
-         * is to ignore & try to continue */
-	open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-	open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-	open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+	gdm_close_all_descriptors (3 /* from */, -1 /* except */);
 
 	openlog ("gdm", LOG_PID, LOG_DAEMON);
 	
@@ -2293,21 +2360,6 @@ session_child_run (struct passwd *pwent,
 		shell = "/bin/sh";
 	}
 
-        /* Log all output from session programs to a file,
-	 * unless in failsafe mode which needs to work when there is
-	 * no diskspace as well */
-	if ( ! failsafe && home_dir_ok) {
-		logfd = open (g_strconcat (home_dir, "/.xsession-errors", NULL),
-			      O_CREAT|O_TRUNC|O_WRONLY, 0644);
-		if (logfd != -1) {
-			dup2 (logfd, 1);
-			dup2 (logfd, 2);
-		} else {
-			gdm_error (_("%s: Could not open ~/.xsession-errors"),
-				   "run_session_child");
-		}
-	}
-
 	/* just a stupid test, the below would fail, but this gives a better
 	 * message */
 	if (strcmp (shell, "/sbin/nologin") == 0 ||
@@ -2329,7 +2381,33 @@ session_child_run (struct passwd *pwent,
 			   "dialog window."));
 	} else {
 		char *exec = g_strconcat ("exec ", sesspath, NULL);
+		char *shellbase = g_path_get_basename (shell);
+		/* FIXME: this is a hack currently this is all screwed up
+		 * so I won't run the users shell unless it's one of the
+		 * "listed" ones, I'll just run bash or sh,
+		 * that's a bit evil but in the end it works out better in
+		 * fact.  In the future we will do our own login setup
+		 * stuff */
+		if (strcmp (shellbase, "sh") != 0 &&
+		    strcmp (shellbase, "bash") == 0 &&
+		    strcmp (shellbase, "tcsh") == 0 &&
+		    strcmp (shellbase, "ksh") == 0 &&
+		    strcmp (shellbase, "pdksh") == 0 &&
+		    strcmp (shellbase, "zsh") == 0 &&
+		    strcmp (shellbase, "csh") == 0 &&
+		    strcmp (shellbase, "ash") == 0 &&
+		    strcmp (shellbase, "bsh") == 0 &&
+		    strcmp (shellbase, "bash2") == 0) {
+			if (access ("/bin/bash", R_OK|X_OK) == 0)
+				shell = "/bin/bash";
+			else if (access ("/bin/bash2", R_OK|X_OK) == 0)
+				shell = "/bin/bash2";
+			else
+				shell = "/bin/sh";
+		}
 		execl (shell, "-", "-c", exec, NULL);
+		/* nutcase fallback */
+		execl ("/bin/sh", "-", "-c", exec, NULL);
 
 		gdm_error (_("gdm_slave_session_start: Could not start session `%s'"), sesspath);
 		gdm_error_box
@@ -2483,8 +2561,7 @@ gdm_slave_session_start (void)
 
 	    g_free (language);
 
-	    if (lang != NULL &&
-		lang[0] != '\0') 
+	    if ( ! ve_string_empty (lang))
 		    language = g_strdup (lang);
 	    else
 		    language = g_strdup (GdmDefaultLocale);
@@ -2542,15 +2619,6 @@ gdm_slave_session_start (void)
     if (GdmKillInitClients)
 	    gdm_server_whack_clients (d);
 
-    /* setup some env for PreSession script */
-    gnome_setenv ("DISPLAY", d->name, TRUE);
-
-    /* If script fails reset X server and restart greeter */
-    if (gdm_slave_exec_script (d, GdmPreSession,
-			       login, pwent) != EXIT_SUCCESS) 
-	gdm_slave_exit (DISPLAY_REMANAGE,
-			_("gdm_slave_session_start: Execution of PreSession script returned > 0. Aborting."));
-
     /* Setup cookie -- We need this information during cleanup, thus
      * cookie handling is done before fork()ing */
 
@@ -2590,9 +2658,12 @@ gdm_slave_session_start (void)
 
     if (strcmp (session, GDM_SESSION_FAILSAFE_GNOME) == 0 ||
 	strcmp (session, GDM_SESSION_FAILSAFE_XTERM) == 0 ||
-	strcmp (session, "Failsafe") == 0 /* hack */)
+	g_ascii_strcasecmp (session, "Failsafe") == 0 /* hack */)
 	    failsafe = TRUE;
 
+    /* Write out the Xservers file */
+    gdm_slave_send_num (GDM_SOP_WRITE_X_SERVERS, 0 /* bogus */);
+
     /* don't completely rely on this, the user
      * could reset time or do other crazy things */
     session_start_time = time (NULL);
@@ -2610,7 +2681,6 @@ gdm_slave_session_start (void)
 	gdm_slave_exit (DISPLAY_ABORT, _("gdm_slave_session_start: Error forking user session"));
 	
     case 0:
-
 	/* Never returns */
 	session_child_run (pwent,
 			   home_dir,
@@ -2642,11 +2712,11 @@ gdm_slave_session_start (void)
     if (d->sesspid > 0) {
 	    gdm_slave_send_num (GDM_SOP_SESSPID, pid);
 
-	    d->sesspid = 0;
-
 	    gdm_sigchld_block_pop ();
 
-	    waitpid (pid, NULL, 0);
+	    ve_waitpid_no_signal (pid, NULL, 0);
+
+	    d->sesspid = 0;
     } else {
 	    gdm_sigchld_block_pop ();
     }
@@ -2692,6 +2762,7 @@ static void
 gdm_slave_session_stop (pid_t sesspid)
 {
     struct passwd *pwent;
+    char *x_servers_file;
     char *local_login;
 
     local_login = login;
@@ -2704,16 +2775,44 @@ gdm_slave_session_stop (pid_t sesspid)
 
     gdm_debug ("gdm_slave_session_stop: %s on %s", local_login, d->name);
 
-    if (sesspid > 0)
-	    kill (- (sesspid), SIGTERM);
+    /* Note we use the info in the structure here since if we get passed
+     * a 0 that means the process is already dead.
+     * FIXME: Maybe we should waitpid here, note make sure this will
+     * not create a hang! */
+    gdm_sigchld_block_push ();
+    if (d->sesspid > 0)
+	    kill (- (d->sesspid), SIGTERM);
+    gdm_sigchld_block_pop ();
 
     gdm_verify_cleanup (d);
     
     pwent = getpwnam (local_login);	/* PAM overwrites our pwent */
 
-    /* Execute post session script */
-    gdm_debug ("gdm_slave_session_cleanup: Running post session script");
-    gdm_slave_exec_script (d, GdmPostSession, local_login, pwent);
+    x_servers_file = g_strconcat (GdmServAuthDir,
+				  "/", d->name, ".Xservers", NULL);
+
+    /* if there was a session that ran, run the PostSession script */
+    if (sesspid != 0) {
+	    /* setup some env for PostSession script */
+	    gnome_setenv ("DISPLAY", d->name, TRUE);
+	    gnome_setenv ("XAUTHORITY", d->authfile, TRUE);
+
+	    gnome_setenv ("X_SERVERS", x_servers_file, TRUE);
+	    if (d->type == TYPE_XDMCP)
+		    gnome_setenv ("REMOTE_HOST", d->hostname, TRUE);
+
+	    /* Execute post session script */
+	    gdm_debug ("gdm_slave_session_cleanup: Running post session script");
+	    gdm_slave_exec_script (d, GdmPostSession, local_login, pwent,
+				   FALSE /* pass_stdout */);
+
+	    gnome_unsetenv ("X_SERVERS");
+	    if (d->type == TYPE_XDMCP)
+		    gnome_unsetenv ("REMOTE_HOST");
+    }
+
+    unlink (x_servers_file);
+    g_free (x_servers_file);
 
     if (pwent == NULL) {
 	    return;
@@ -2781,7 +2880,7 @@ gdm_slave_term_handler (int sig)
 		greet = FALSE;
 		gdm_debug ("gdm_slave_term_handler: Whacking greeter");
 		if (kill (pid, sig) == 0)
-			waitpid (pid, 0, 0); 
+			ve_waitpid_no_signal (pid, 0, 0); 
 	} else if (login != NULL) {
 		gdm_slave_session_stop (d->sesspid);
 		gdm_slave_session_cleanup ();
@@ -2792,7 +2891,7 @@ gdm_slave_term_handler (int sig)
 		d->chooserpid = 0;
 		gdm_debug ("gdm_slave_term_handler: Whacking chooser");
 		if (kill (pid, sig) == 0)
-			waitpid (pid, 0, 0); 
+			ve_waitpid_no_signal (pid, 0, 0); 
 	}
 
 	gdm_debug ("gdm_slave_term_handler: Whacking server");
@@ -2999,7 +3098,7 @@ gdm_slave_xioerror_handler (Display *disp)
 		d->greetpid = 0;
 		greet = FALSE;
 		if (kill (pid, SIGINT) == 0)
-			waitpid (pid, 0, 0); 
+			ve_waitpid_no_signal (pid, 0, 0); 
 	} else if (login != NULL) {
 		gdm_slave_session_stop (d->sesspid);
 		gdm_slave_session_cleanup ();
@@ -3009,7 +3108,7 @@ gdm_slave_xioerror_handler (Display *disp)
 		pid_t pid = d->chooserpid;
 		d->chooserpid = 0;
 		if (kill (pid, SIGINT) == 0)
-			waitpid (pid, 0, 0); 
+			ve_waitpid_no_signal (pid, 0, 0); 
 	}
     
 	gdm_error (_("gdm_slave_xioerror_handler: Fatal X error - Restarting %s"), d->name);
@@ -3079,7 +3178,7 @@ check_for_interruption (const char *msg)
 char * 
 gdm_slave_greeter_ctl (char cmd, const char *str)
 {
-    gchar buf[FIELD_SIZE];
+    char *buf = NULL;
     int c;
 
     /* There is no spoon^H^H^H^H^Hgreeter */
@@ -3090,18 +3189,18 @@ gdm_slave_greeter_ctl (char cmd, const char *str)
     check_notifies_immediately ++;
 
     if ( ! ve_string_empty (str)) {
-	    printf ("%c%c%s\n", STX, cmd, str);
+	    gdm_fdprintf (STDOUT_FILENO, "%c%c%s\n", STX, cmd, str);
     } else {
-	    printf ("%c%c\n", STX, cmd);
+	    gdm_fdprintf (STDOUT_FILENO, "%c%c\n", STX, cmd);
     }
-    fflush (stdout);
 
     /* Skip random junk that might have accumulated */
     do {
-	    c = getc (stdin);
+	    c = gdm_fdgetc (STDIN_FILENO);
     } while (c != EOF && c != STX);
     
-    if (c == EOF || fgets (buf, FIELD_SIZE-1, stdin) == NULL) {
+    if (c == EOF ||
+	(buf = gdm_fdgets (STDIN_FILENO)) == NULL) {
 	    check_notifies_immediately --;
 	    interrupted = TRUE;
 	    /* things don't seem well with the greeter, it probably died */
@@ -3112,15 +3211,10 @@ gdm_slave_greeter_ctl (char cmd, const char *str)
 
     check_for_interruption (buf);
 
-    /* don't forget to flush */
-    fflush (stdin);
-    
-    if (buf[0] != '\0') {
-	    int len = strlen (buf);
-	    if (buf[len-1] == '\n')
-		    buf[len-1] = '\0';
-	    return g_strdup (buf);
+    if ( ! ve_string_empty (buf)) {
+	    return buf;
     } else {
+	    g_free (buf);
 	    return NULL;
     }
 }
@@ -3200,14 +3294,14 @@ gdm_child_exit (gint status, const gchar *format, ...)
 
 static gint
 gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
-		       struct passwd *pwent)
+		       struct passwd *pwent, gboolean pass_stdout)
 {
     pid_t pid;
     gchar *script, *defscript;
     const char *scr;
     gchar **argv;
     gint status;
-    int i;
+    char *x_servers_file;
 
     if (!d || !dir)
 	return EXIT_SUCCESS;
@@ -3215,9 +3309,9 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
     script = g_strconcat (dir, "/", d->name, NULL);
     defscript = g_strconcat (dir, "/Default", NULL);
 
-    if (! access (script, R_OK|X_OK)) {
+    if (access (script, R_OK|X_OK) == 0) {
 	    scr = script;
-    } else if (! access (defscript, R_OK|X_OK))  {
+    } else if (access (defscript, R_OK|X_OK) == 0)  {
 	    scr = defscript;
     } else {
 	    g_free (script);
@@ -3232,14 +3326,19 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
     case 0:
         closelog ();
 
-        for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-	    close (i);
+	close (0);
+	open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
 
-        /* No error checking here - if it's messed the best response
-         * is to ignore & try to continue */
-        open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-        open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-        open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+	if ( ! pass_stdout) {
+		close (1);
+		close (2);
+		/* No error checking here - if it's messed the best response
+		 * is to ignore & try to continue */
+		open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
+		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+	}
+
+	gdm_close_all_descriptors (3 /* from */, -1 /* except */);
 
 	openlog ("gdm", LOG_PID, LOG_DAEMON);
 
@@ -3263,6 +3362,14 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
 	        gnome_setenv ("SHELL", "/bin/sh", TRUE);
         }
 
+	/* some env for use with the Pre and Post scripts */
+	x_servers_file = g_strconcat (GdmServAuthDir,
+				      "/", d->name, ".Xservers", NULL);
+	gnome_setenv ("X_SERVERS", x_servers_file, TRUE);
+	g_free (x_servers_file);
+	if (d->type == TYPE_XDMCP)
+		gnome_setenv ("REMOTE_HOST", d->hostname, TRUE);
+
 	/* Runs as root, uses server authfile */
 	gnome_setenv ("XAUTHORITY", d->authfile, TRUE);
         gnome_setenv ("DISPLAY", d->name, TRUE);
@@ -3371,7 +3478,6 @@ gdm_parse_enriched_login (const gchar *s, GdmDisplay *display)
       if (pipe (pipe1) < 0) {
         gdm_error (_("gdm_parse_enriched_login: Failed creating pipe"));
       } else {
-        int i;
 	pid = gdm_fork_extra ();
 
         switch (pid) {
@@ -3386,10 +3492,7 @@ gdm_parse_enriched_login (const gchar *s, GdmDisplay *display)
 
 	    closelog ();
 
-	    for (i = 3; i < sysconf (_SC_OPEN_MAX); i++) {
-		    if (pipe1[1] != i)
-			    close (i);
-	    }
+	    gdm_close_all_descriptors (3 /* from */, pipe1[1] /* except */);
 
 	    openlog ("gdm", LOG_PID, LOG_DAEMON);
 
