@@ -1,7 +1,7 @@
 /* Misc routines
  *
  * (c) 2000 Eazel, Inc.
- * (c) 2001 George Lebl
+ * (c) 2001,2002 George Lebl
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -22,6 +22,8 @@
 #define VE_MISC_H
 
 #include <time.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 
 char **	ve_split (const char *s);
 char *	ve_first_word (const char *s);
@@ -54,4 +56,9 @@ char * ve_locale_from_utf8 (const char *str);
 /* works with utf-8 strings and always returns something */
 char * ve_strftime (struct tm *the_tm, const char *format);
 
+/* function which doesn't stop on signals */
+pid_t ve_waitpid_no_signal (pid_t pid, int *status, int options);
+
+
+
 #endif /* VE_MISC_H */
